import inspect

import matplotlib.pyplot as plt
import numpy as np
from scipy.interpolate import interp1d
from scipy.optimize import curve_fit
from scipy.stats import linregress
from sklearn.linear_model import LinearRegression

from . import basis, kp, kt

__author__ = "Jieqing Jiao <jieqing.jiao@gmail.com>"
# excludes:
# - exp_am: input interpolation requested by some users
# - *_ppet: kinetic model implementations of the PPET software
# - *_para2tac: generative models for motion correction
NAMES = [
    'exp_1', 'exp_1_fun', 'exp_1_fun_t', 'exp_2', 'exp_2_fun', 'exp_2_fun_t', 'feng_fun_t',
    'feng_srtm', 'feng_srtm_fun', 'feng_srtm_fun_t', 'list_models', 'logan_ref', 'logan_ref_k2p',
    'mrtm', 'mrtm_k2p', 'srtm', 'srtm_fun', 'srtm_fun_k2p', 'srtm_fun_k2p_w', 'srtm_fun_w',
    'srtm_k2p', 'srtmb', 'srtmb_asl', 'srtmb_asl_basis', 'srtmb_basis', 'srtmb_k2p',
    'srtmb_k2p_basis']
__all__ = NAMES + ['NAMES', 'get_model_inputs']


def get_model_inputs(user_inputs, model_name):
    """select model args from user_inputs"""
    if model_name not in NAMES:
        raise ValueError(f"Unknown model:'{model_name}' not in {NAMES}")
    sig = inspect.signature(globals()[model_name])
    model_inputs = {}
    for p in sig.parameters.values():
        n = p.name
        # d = None if p.default == inspect.Parameter.empty else p.default
        if n in user_inputs:
            model_inputs[n] = user_inputs[n]
    return model_inputs


# # # # # # # # # # # # # # # # # # # # # # # # # # # #
# # linear models
# # # # # # # # # # # # # # # # # # # # # # # # # # # #


def srtmb_basis(tac, b):
    """srtm model for img with pre-calculated basis functions"""
    tac[tac < 0] = 0.0

    n_beta = b['beta'].size
    ssq = np.zeros(n_beta)

    if b['w'] is None:
        b['w'] = 1

    for i in range(0, n_beta):
        theta = np.dot(b['m_w'][i * 2:i*2 + 2, :], b['w'] * tac)
        a = np.column_stack((b['input'], b['basis'][i, :]))
        tacf = np.dot(a, theta)
        res = (tac-tacf) * b['w']
        ssq[i] = np.sum(res**2)
    i = np.argmin(ssq)
    theta = np.dot(b['m_w'][i * 2:i*2 + 2, :], b['w'] * tac)
    a = np.column_stack((b['input'], b['basis'][i, :]))
    tacf = np.dot(a, theta)

    theta = np.append(theta, b['beta'][i])
    r1, k2, bp = kp.srtm_theta2kp(theta)
    kps = {'r1': r1, 'k2': k2, 'bp': bp, 'tacf': tacf}
    return kps


def srtmb_basis_para2tac(r1, k2, bp, b):
    tacf = []
    theta = kp.srtm_kp2theta(r1, k2, bp)
    i = np.argwhere(abs(b['beta'] - theta[-1]) < 1e-10).squeeze()
    if not i == []:
        a = np.column_stack((b['input'], b['basis'][i, :]))
        tacf = np.dot(a, theta[:-1])
    kps = {'tacf': tacf}
    return kps


def srtmb(tac, dt, inputf1, beta_lim, n_beta, w):
    """srtm model for tac, basis functions will be calculated"""
    tac[tac < 0] = 0.0
    b = basis.make_basis(inputf1, dt, beta_lim=beta_lim, n_beta=n_beta, w=w)
    return srtmb_basis(tac, b)


def srtmb_asl_basis(tac, b, r1):
    """srtm model for tac with fixed R1 and pre-calculated basis functions"""
    tac[tac < 0] = 0.0

    n_beta = b['beta'].size
    ssq = np.zeros(n_beta)

    if b['w'] is None:
        b['w'] = 1
    y = tac - r1 * b['input']
    for i in range(0, n_beta):
        theta = r1
        theta = np.append(
            theta,
            np.dot(y, b['basis'][i, :]) /
            np.dot(b['basis'][i, :], b['basis'][i, :])) # w won't make a difference here
        a = np.column_stack((b['input'], b['basis'][i, :]))
        tacf = np.dot(a, theta)
        res = (tac-tacf) * b['w']                       # w works here
        ssq[i] = np.sum(res**2)
    i = np.argmin(ssq)
    theta = r1
    theta = np.append(theta,
                      np.dot(y, b['basis'][i, :]) / np.dot(b['basis'][i, :], b['basis'][i, :]))
    a = np.column_stack((b['input'], b['basis'][i, :]))
    tacf = np.dot(a, theta)

    theta = np.append(theta, b['beta'][i])
    r1, k2, bp = kp.srtm_theta2kp(theta)
    kps = {'r1': r1, 'k2': k2, 'bp': bp, 'tacf': tacf}
    return kps


def srtmb_asl(tac, dt, inputf1, beta_lim, n_beta, w, r1):
    """srtm model for tac with fixed R1, basis functions will be calculated"""
    tac[tac < 0] = 0.0
    b = basis.make_basis(inputf1, dt, beta_lim=beta_lim, n_beta=n_beta, w=w)
    kps = srtmb_asl_basis(tac, b, r1)
    return kps


def srtmb_k2p_basis(tac, b):
    """srtm model for img with fixed k2p and pre-calculated basis functions"""
    tac[tac < 0] = 0.0

    n_beta = b['beta'].size
    ssq = np.zeros(n_beta)

    if b['w'] is None:
        b['w'] = 1
    for i in range(0, n_beta):
        r1 = np.sum(b['w'] * b['basis_k2p'][i] * tac) / np.sum(b['w'] * b['basis_k2p'][i]**2)
        ssq[i] = np.sum(b['w'] * (tac - r1 * b['basis_k2p'][i])**2)

    i = np.argmin(ssq)
    r1 = np.sum(b['w'] * b['basis_k2p'][i] * tac) / np.sum(b['w'] * b['basis_k2p'][i]**2)
    tacf = r1 * b['basis_k2p'][i]

    theta = r1
    theta = np.append(theta, r1 * (b['k2p'] - b['beta'][i]))
    theta = np.append(theta, b['beta'][i])
    r1, k2, bp = kp.srtm_theta2kp(theta)
    kps = {'r1': r1, 'k2': k2, 'bp': bp, 'tacf': tacf}
    return kps


def srtmb_k2p_basis_para2tac(r1, k2, bp, b):
    tacf = []
    theta = kp.srtm_kp2theta(r1, k2, bp)
    i = np.argwhere(abs(b['beta'] - theta[-1]) < 1e-10).squeeze()
    if not i == []:
        tacf = theta[0] * b['basis_k2p'][i]
    kps = {'tacf': tacf}
    return kps


def srtmb_k2p(tac, dt, inputf1, beta_lim, n_beta, w, k2p):
    """srtm model for tac with fixed k2p, basis functions will be calculated"""
    tac[tac < 0] = 0.0
    b = basis.make_basis(inputf1, dt, beta_lim=beta_lim, n_beta=n_beta, w=w, k2p=k2p)
    kps = srtmb_k2p_basis(tac, b)
    return kps


# # # # # # graphic models


def logan_ref(tac, dt, inputf1, linear_phase_start, linear_phase_end, fig):
    """
    logan reference plot without fixed k2p for tac, based on eq.7 in
    "Distribution Volume Ratios Without Blood Sampling from Graphical Analysis of PET Data"
    """
    if linear_phase_start is None:
        linear_phase_start = 0
    if linear_phase_end is None:
        linear_phase_end = np.amax(dt)
    # fill the coffee break gap
    if kt.dt_has_gaps(dt):
        tac, dt = kt.tac_dt_fill_coffee_break(tac, dt)
    mft = kt.dt2mft(dt)
    mft = np.append(0, mft)
    dt_new = np.array([mft[:-1], mft[1:]])
    tdur = kt.dt2tdur(dt_new)
    # input_dt = kt.int2dt(inputf1, dt)
    inputff = interp1d(np.arange(len(inputf1)), inputf1, kind='linear', fill_value='extrapolate')
    input_dt = inputff(mft)
    input_dt = input_dt[1:]

    tac = np.append(0, tac)
    input_dt = np.append(0, input_dt)

    # set negative values to zero
    tac[tac < 0] = 0.0
    input_dt[input_dt < 0] = 0.0

    # calculate integration
    tac_cum = np.cumsum((tac[:-1] + tac[1:]) / 2 * tdur)
    input_cum = np.cumsum((input_dt[:-1] + input_dt[1:]) / 2 * tdur)
    tac = tac[1:]
    input_dt = input_dt[1:]

    yy = tac_cum / (tac+0.0000000000000001)   # ADDED BY MY 20210616
    xx = input_cum / (tac+0.0000000000000001) # ADDED BY MY 20210616

    # find tt for the linear phase
    tt = np.logical_and(mft >= linear_phase_start, mft <= linear_phase_end)
    tt = tt[1:]
    # select tt for tac > 0
    tt = np.logical_and(tt, tac > 0)
    # select tt for xx < inf, yy < inf
    infinf = 1e10
    tt = np.logical_and(tt, xx < infinf)
    tt = np.logical_and(tt, yy < infinf)

    # do linear regression with selected tt
    xx = xx[tt]
    yy = yy[tt]

    dvr, inter, _, _, _ = linregress(xx, yy)
    bp = dvr - 1
    yyf = dvr*xx + inter
    if fig:
        plt.plot(xx, yy, '.')
        plt.plot(xx, yyf, 'r')
        plt.show()
    kps = {'bp': bp}
    return kps


def logan_ref_k2p(tac, dt, inputf1, k2p, linear_phase_start, linear_phase_end, fig):
    """
    logan reference plot with fixed k2p for tac, based on eq.6 in
    "Distribution Volume Ratios Without Blood Sampling from Graphical Analysis of PET Data"
    """
    if linear_phase_start is None:
        linear_phase_start = 0
    if linear_phase_end is None:
        linear_phase_end = np.amax(dt)
    # fill the coffee break gap
    if kt.dt_has_gaps(dt):
        tac, dt = kt.tac_dt_fill_coffee_break(tac, dt)
    mft = kt.dt2mft(dt)
    mft = np.append(0, mft)
    dt_new = np.array([mft[:-1], mft[1:]])
    tdur = kt.dt2tdur(dt_new)
    # input_dt = kt.int2dt(inputf1, dt)
    inputff = interp1d(np.arange(len(inputf1)), inputf1, kind='linear', fill_value='extrapolate')
    input_dt = inputff(mft)
    input_dt = input_dt[1:]
    tac = np.append(0, tac)
    input_dt = np.append(0, input_dt)

    # set negative values to zero
    tac[tac < 0] = 0.0
    input_dt[input_dt < 0] = 0.0

    tac_cum = np.cumsum((tac[:-1] + tac[1:]) / 2 * tdur)
    input_cum = np.cumsum((input_dt[:-1] + input_dt[1:]) / 2 * tdur)
    tac = tac[1:]
    input_dt = input_dt[1:]

    yy = tac_cum / (tac+0.0000000000000001)
    xx = (input_cum + input_dt/k2p) / (tac+0.0000000000000001)

    # find tt for the linear phase
    tt = np.logical_and(mft >= linear_phase_start, mft <= linear_phase_end)
    tt = tt[1:]
    # select tt for tac > 0
    tt = np.logical_and(tt, tac > 0)
    # select tt for xx < inf, yy < inf
    infinf = 1e10
    tt = np.logical_and(tt, xx < infinf)
    tt = np.logical_and(tt, yy < infinf)

    # do linear regression with selected tt
    xx = xx[tt]
    yy = yy[tt]

    dvr, inter, _, _, _ = linregress(xx, yy)
    bp = dvr - 1
    yyf = dvr*xx + inter
    if fig:
        plt.plot(xx, yy, '.')
        plt.plot(xx[tt], yyf[tt], 'r')
        plt.show()
    kps = {'bp': bp}
    return kps


def mrtm(tac, dt, inputf1, linear_phase_start, linear_phase_end, fig):
    """Ichise's multilinear reference tissue model"""
    if linear_phase_start is None:
        linear_phase_start = 0
    if linear_phase_end is None:
        linear_phase_end = np.amax(dt)
    # fill the coffee break gap
    if kt.dt_has_gaps(dt):
        tac, dt = kt.tac_dt_fill_coffee_break(tac, dt)
    mft = kt.dt2mft(dt)
    mft = np.append(0, mft)
    dt_new = np.array([mft[:-1], mft[1:]])
    tdur = kt.dt2tdur(dt_new)
    # input_dt = kt.int2dt(inputf1, dt)
    inputff = interp1d(np.arange(len(inputf1)), inputf1, kind='linear', fill_value='extrapolate')
    input_dt = inputff(mft)
    input_dt = input_dt[1:]
    tac = np.append(0, tac)
    input_dt = np.append(0, input_dt)

    # set negative values to zero
    tac[tac < 0] = 0.0
    input_dt[input_dt < 0] = 0.0

    tac_cum = np.cumsum((tac[:-1] + tac[1:]) / 2 * tdur)
    input_cum = np.cumsum((input_dt[:-1] + input_dt[1:]) / 2 * tdur)
    tac = tac[1:]
    input_dt = input_dt[1:]
    yy = tac
    xx = np.column_stack((input_cum, tac_cum, input_dt))

    # find tt for the linear phase
    tt = np.logical_and(mft >= linear_phase_start, mft <= linear_phase_end)
    tt = tt[1:]

    mft = mft[1:]

    reg = LinearRegression(fit_intercept=False).fit(xx[tt,], yy[tt])
    bp = -reg.coef_[0] / reg.coef_[1] - 1
    k2p = reg.coef_[0] / reg.coef_[2]
    # for 1 TC
    r1 = reg.coef_[2]
    k2 = -reg.coef_[1]

    if np.isnan(bp):
        bp = 0
    if np.isnan(r1):
        r1 = 1.0
    if r1 > 5:
        r1 = 5
    if r1 < -5:
        r1 = -5
    if bp > 10:
        bp = 0
    if bp < -10:
        bp = 0

    yyf = reg.predict(xx)
    if fig:
        plt.plot(mft, yy, '.')
        plt.plot(mft, yyf, 'r')
        plt.show()
    kps = {'bp': bp, 'k2p': k2p, 'r1': r1, 'k2': k2}
    return kps


def mrtm_k2p(tac, dt, inputf1, k2p, linear_phase_start, linear_phase_end, fig):
    """Ichise's multilinear reference tissue model with fixed k2prime"""
    if linear_phase_start is None:
        linear_phase_start = 0
    if linear_phase_end is None:
        linear_phase_end = np.amax(dt)
    # fill the coffee break gap
    if kt.dt_has_gaps(dt):
        tac, dt = kt.tac_dt_fill_coffee_break(tac, dt)
    mft = kt.dt2mft(dt)
    mft = np.append(0, mft)
    dt_new = np.array([mft[:-1], mft[1:]])
    tdur = kt.dt2tdur(dt_new)
    # input_dt = kt.int2dt(inputf1,dt)
    inputff = interp1d(np.arange(len(inputf1)), inputf1, kind='linear', fill_value='extrapolate')
    input_dt = inputff(mft)
    input_dt = input_dt[1:]
    tac = np.append(0, tac)
    input_dt = np.append(0, input_dt)

    # set negative values to zero
    tac[tac < 0] = 0.0
    input_dt[input_dt < 0] = 0.0

    tac_cum = np.cumsum((tac[:-1] + tac[1:]) / 2 * tdur)
    input_cum = np.cumsum((input_dt[:-1] + input_dt[1:]) / 2 * tdur)
    tac = tac[1:]
    input_dt = input_dt[1:]

    yy = tac
    xx = np.column_stack((input_cum + 1/k2p*input_dt, tac_cum))

    # find tt for the linear phase
    tt = np.logical_and(mft >= linear_phase_start, mft <= linear_phase_end)
    tt = tt[1:]

    mft = mft[1:]
    reg = LinearRegression(fit_intercept=False).fit(xx[tt,], yy[tt])
    bp = -reg.coef_[0] / reg.coef_[1] - 1

    # for 1 TC
    # k2 = -reg.coef_[1]
    r1 = reg.coef_[0] / k2p

    if np.isnan(bp):
        bp = 0
    if np.isnan(r1):
        r1 = 1.0
    if r1 > 5:
        r1 = 5
    if r1 < -5:
        r1 = -5
    if bp > 10:
        bp = 0
    if bp < -10:
        bp = 0

    yyf = reg.predict(xx)
    if fig:
        plt.plot(mft, yy, '.')
        plt.plot(mft, yyf, 'r')
        plt.show()
    kps = {'bp': bp, 'r1': r1}
    return kps


# # # # # # graphic models - PPET version


def logan_ref_ppet(tac, dt, ref, linear_phase_start, linear_phase_end, fig):
    """
    logan_ref - logan reference plot without fixed k2p for tac, based on eq.7 in
    "Distribution Volume Ratios Without Blood Sampling from Graphical Analysis of PET Data"
    PPET version: calculate input_dt and input_cum differently
    """
    if linear_phase_start is None:
        linear_phase_start = 0
    if linear_phase_end is None:
        linear_phase_end = np.amax(dt)
    # fill the coffee break gap
    if kt.dt_has_gaps(dt):
        tac, dt = kt.tac_dt_fill_coffee_break(tac, dt)
    mft = kt.dt2mft(dt)
    mft = np.append(0, mft)
    dt_new = np.array([mft[:-1], mft[1:]])
    tdur = kt.dt2tdur(dt_new)
    # get input_dt from ref.tac, if dt and ref.dt are the same
    if np.array_equal(dt, ref.dt):
        input_dt = ref.tac
    else:
        inputff = interp1d(kt.dt2mft(ref.dt), ref.tac, kind='linear', fill_value='extrapolate')
        input_dt = inputff(mft)
        input_dt = input_dt[1:]

    tac = np.append(0, tac)
    input_dt = np.append(0, input_dt)

    # set negative values to zero
    tac[tac < 0] = 0.0
    input_dt[input_dt < 0] = 0.0

    # calculate integration
    tac_cum = np.cumsum((tac[:-1] + tac[1:]) / 2 * tdur)
    # calculate input_cum using inputf1, and only until mid frame time
    inputf1 = kt.interpt1(kt.dt2mft(ref.dt), ref.tac, dt)
    input_cum1 = np.cumsum(inputf1)
    input_cum1_mft = input_cum1[mft.astype(int)]
    input_cum = input_cum1_mft
    # tac_cum and input_cum are calculated in such a way to match tac

    tac = tac[1:]
    input_dt = input_dt[1:]
    input_cum = input_cum[1:]

    yy = tac_cum / (tac+0.0000000000000001)   # ADDED BY MY 20210616
    xx = input_cum / (tac+0.0000000000000001) # ADDED BY MY 20210616

    # find tt for the linear phase
    tt = np.logical_and(mft >= linear_phase_start, mft <= linear_phase_end)
    tt = tt[1:]
    # select tt for tac > 0
    tt = np.logical_and(tt, tac > 0)
    # select tt for xx < inf, yy < inf
    infinf = 1e10
    tt = np.logical_and(tt, xx < infinf)
    tt = np.logical_and(tt, yy < infinf)

    # do linear regression with selected tt
    xx = xx[tt]
    yy = yy[tt]

    dvr, inter, _, _, _ = linregress(xx, yy)
    bp = dvr - 1
    yyf = dvr*xx + inter
    if fig:
        plt.plot(xx, yy, '.')
        plt.plot(xx, yyf, 'r')
        plt.show()
    kps = {'bp': bp}
    return kps


def logan_ref_k2p_ppet(tac, dt, ref, k2p, linear_phase_start, linear_phase_end, fig):
    """
    logan_ref_k2p - logan reference plot with fixed k2p for tac, based on eq.6 in
    "Distribution Volume Ratios Without Blood Sampling from Graphical Analysis of PET Data"
    PPET version: calculate input_dt and input_cum differently
    """
    if linear_phase_start is None:
        linear_phase_start = 0
    if linear_phase_end is None:
        linear_phase_end = np.amax(dt)
    # fill the coffee break gap
    if kt.dt_has_gaps(dt):
        tac, dt = kt.tac_dt_fill_coffee_break(tac, dt)
    mft = kt.dt2mft(dt)
    mft = np.append(0, mft)
    dt_new = np.array([mft[:-1], mft[1:]])
    tdur = kt.dt2tdur(dt_new)
    # get input_dt from ref.tac, if dt and ref.dt are the same
    if np.array_equal(dt, ref.dt):
        input_dt = ref.tac
    else:
        inputff = interp1d(kt.dt2mft(ref.dt), ref.tac, kind='linear', fill_value='extrapolate')
        input_dt = inputff(mft)
        input_dt = input_dt[1:]

    tac = np.append(0, tac)
    input_dt = np.append(0, input_dt)

    # set negative values to zero
    tac[tac < 0] = 0.0
    input_dt[input_dt < 0] = 0.0

    tac_cum = np.cumsum((tac[:-1] + tac[1:]) / 2 * tdur)
    # calculate input_cum using inputf1, and only until mid frame time
    inputf1 = kt.interpt1(kt.dt2mft(ref.dt), ref.tac, dt)
    input_cum1 = np.cumsum(inputf1)
    input_cum1_mft = input_cum1[mft.astype(int)]
    input_cum = input_cum1_mft
    # tac_cum and input_cum are calculated in such a way to match tac
    tac = tac[1:]
    input_dt = input_dt[1:]
    input_cum = input_cum[1:]

    yy = tac_cum / (tac+0.0000000000000001)
    xx = (input_cum + input_dt/k2p) / (tac+0.0000000000000001)

    # find tt for the linear phase
    tt = np.logical_and(mft >= linear_phase_start, mft <= linear_phase_end)
    tt = tt[1:]
    # select tt for tac > 0
    tt = np.logical_and(tt, tac > 0)
    # select tt for xx < inf, yy < inf
    infinf = 1e10
    tt = np.logical_and(tt, xx < infinf)
    tt = np.logical_and(tt, yy < infinf)

    # do linear regression with selected tt
    xx = xx[tt]
    yy = yy[tt]

    dvr, inter, _, _, _ = linregress(xx, yy)
    bp = dvr - 1
    yyf = dvr*xx + inter
    if fig:
        plt.plot(xx, yy, '.')
        plt.plot(xx[tt], yyf[tt], 'r')
        plt.show()
    kps = {'bp': bp}
    return kps


def mrtm_ppet(tac, dt, ref, linear_phase_start, linear_phase_end, fig):
    """
    Ichise's multilinear reference tissue model.
    PPET version: calculate input_dt and input_cum differently
    """
    if linear_phase_start is None:
        linear_phase_start = 0
    if linear_phase_end is None:
        linear_phase_end = np.amax(dt)
    # fill the coffee break gap
    if kt.dt_has_gaps(dt):
        tac, dt = kt.tac_dt_fill_coffee_break(tac, dt)
    mft = kt.dt2mft(dt)
    mft = np.append(0, mft)
    dt_new = np.array([mft[:-1], mft[1:]])
    tdur = kt.dt2tdur(dt_new)
    # get input_dt from ref.tac, if dt and ref.dt are the same
    if np.array_equal(dt, ref.dt):
        input_dt = ref.tac
    else:
        inputff = interp1d(kt.dt2mft(ref.dt), ref.tac, kind='linear', fill_value='extrapolate')
        input_dt = inputff(mft)
        input_dt = input_dt[1:]

    tac = np.append(0, tac)
    input_dt = np.append(0, input_dt)

    # set negative values to zero
    tac[tac < 0] = 0.0
    input_dt[input_dt < 0] = 0.0

    tac_cum = np.cumsum((tac[:-1] + tac[1:]) / 2 * tdur)
    # calculate input_cum using inputf1, and only until mid frame time
    inputf1 = kt.interpt1(kt.dt2mft(ref.dt), ref.tac, dt)
    input_cum1 = np.cumsum(inputf1)
    input_cum1_mft = input_cum1[mft.astype(int)]
    input_cum = input_cum1_mft
    # tac_cum and input_cum are calculated in such a way to match tac
    tac = tac[1:]
    input_dt = input_dt[1:]
    input_cum = input_cum[1:]
    yy = tac
    xx = np.column_stack((input_cum, tac_cum, input_dt))

    # find tt for the linear phase
    tt = np.logical_and(mft >= linear_phase_start, mft <= linear_phase_end)
    tt = tt[1:]

    mft = mft[1:]

    reg = LinearRegression(fit_intercept=False).fit(xx[tt,], yy[tt])
    bp = -reg.coef_[0] / reg.coef_[1] - 1
    k2p = reg.coef_[0] / reg.coef_[2]
    # for 1 TC
    r1 = reg.coef_[2]
    k2 = -reg.coef_[1]

    if np.isnan(bp):
        bp = 0
    if np.isnan(r1):
        r1 = 1.0
    if r1 > 5:
        r1 = 5
    if r1 < -5:
        r1 = -5
    if bp > 10:
        bp = 0
    if bp < -10:
        bp = 0

    yyf = reg.predict(xx)
    if fig:
        plt.plot(mft, yy, '.')
        plt.plot(mft, yyf, 'r')
        plt.show()
    kps = {'bp': bp, 'k2p': k2p, 'r1': r1, 'k2': k2}
    return kps


def mrtm_k2p_ppet(tac, dt, ref, k2p, linear_phase_start, linear_phase_end, fig):
    """
    Ichise's multilinear reference tissue model with fixed k2prime.
    PPET version: calculate input_dt and input_cum differently
    """
    if linear_phase_start is None:
        linear_phase_start = 0
    if linear_phase_end is None:
        linear_phase_end = np.amax(dt)
    # fill the coffee break gap
    if kt.dt_has_gaps(dt):
        tac, dt = kt.tac_dt_fill_coffee_break(tac, dt)
    mft = kt.dt2mft(dt)
    mft = np.append(0, mft)
    dt_new = np.array([mft[:-1], mft[1:]])
    tdur = kt.dt2tdur(dt_new)
    # get input_dt from ref.tac, if dt and ref.dt are the same
    if np.array_equal(dt, ref.dt):
        input_dt = ref.tac
    else:
        inputff = interp1d(kt.dt2mft(ref.dt), ref.tac, kind='linear', fill_value='extrapolate')
        input_dt = inputff(mft)
        input_dt = input_dt[1:]

    tac = np.append(0, tac)
    input_dt = np.append(0, input_dt)

    # set negative values to zero
    tac[tac < 0] = 0.0
    input_dt[input_dt < 0] = 0.0

    tac_cum = np.cumsum((tac[:-1] + tac[1:]) / 2 * tdur)
    # calculate input_cum using inputf1, and only until mid frame time
    inputf1 = kt.interpt1(kt.dt2mft(ref.dt), ref.tac, dt)
    input_cum1 = np.cumsum(inputf1)
    input_cum1_mft = input_cum1[mft.astype(int)]
    input_cum = input_cum1_mft
    # tac_cum and input_cum are calculated in such a way to match tac
    tac = tac[1:]
    input_dt = input_dt[1:]
    input_cum = input_cum[1:]

    yy = tac
    xx = np.column_stack((input_cum + 1/k2p*input_dt, tac_cum))

    # find tt for the linear phase
    tt = np.logical_and(mft >= linear_phase_start, mft <= linear_phase_end)
    tt = tt[1:]

    mft = mft[1:]
    reg = LinearRegression(fit_intercept=False).fit(xx[tt,], yy[tt])
    bp = -reg.coef_[0] / reg.coef_[1] - 1

    # for 1 TC
    # k2 = -reg.coef_[1]
    r1 = reg.coef_[0] / k2p

    if np.isnan(bp):
        bp = 0
    if np.isnan(r1):
        r1 = 1.0
    if r1 > 5:
        r1 = 5
    if r1 < -5:
        r1 = -5
    if bp > 10:
        bp = 0
    if bp < -10:
        bp = 0

    yyf = reg.predict(xx)
    if fig:
        plt.plot(mft, yy, '.')
        plt.plot(mft, yyf, 'r')
        plt.show()
    kps = {'bp': bp, 'r1': r1}
    return kps


# # # # # # # # # # # # # # # # # # # # # # # # # # # #
# # non-linear models
# # # # # # # # # # # # # # # # # # # # # # # # # # # #


def srtm_fun(inputf1_dt, r1, k2, bp):
    """srtm model for tac, with non-linear optimisation"""
    inputf1, dt = inputf1_dt
    t1 = np.arange(np.amax(dt))
    theta = kp.srtm_kp2theta(r1, k2, bp)
    tac1 = theta[0] * inputf1
    tac1 += theta[1] * np.convolve(inputf1, np.exp(-theta[2] * t1))[0:tac1.size]
    tac = kt.int2dt(tac1, dt)
    return tac


def srtm_para2tac(r1, k2, bp, inputf1_dt):
    tacf = srtm_fun(inputf1_dt, r1, k2, bp)
    kps = {'tacf': tacf}
    return kps


def srtm_fun_w(inputf1_dt_w, r1, k2, bp):
    inputf1, dt, w = inputf1_dt_w
    tac = srtm_fun((inputf1, dt), r1, k2, bp)
    if w is None:
        w = 1
    tac_w = tac * w
    return tac_w


def srtm(tac, dt, inputf1, w):
    inputf1_dt = inputf1, dt
    inputf1_dt_w = inputf1, dt, w
    if w is None:
        w = 1
    p, _ = curve_fit(srtm_fun_w, inputf1_dt_w, tac * w, p0=[1, 0.00005, 0.0],
                     bounds=(0, [3, 1, 10]))
    r1 = p[0]
    k2 = p[1]
    bp = p[2]
    tacf = srtm_fun(inputf1_dt, r1, k2, bp)
    kps = {'r1': r1, 'k2': k2, 'bp': bp, 'tacf': tacf}
    return kps


def srtm_fun_k2p(inputf1_dt_k2p, theta_0, theta_2):
    """srtm model for tac with fixed k2p, with non-linear optimisation"""
    inputf1, dt, k2p = inputf1_dt_k2p
    inputf1_dt = (inputf1, dt)
    theta_1 = theta_0 * (k2p-theta_2)
    theta = np.array([theta_0, theta_1, theta_2])
    r1, k2, bp = kp.srtm_theta2kp(theta)
    tac = srtm_fun(inputf1_dt, r1, k2, bp)
    return tac


def srtm_fun_k2p_w(inputf1_dt_k2p_w, theta_0, theta_2):
    inputf1, dt, k2p, w = inputf1_dt_k2p_w
    tac = srtm_fun_k2p((inputf1, dt, k2p), theta_0, theta_2)
    if w is None:
        w = 1
    tac_w = tac * w
    return tac_w


def srtm_k2p(tac, dt, inputf1, w, k2p):
    inputf1_dt_k2p = inputf1, dt, k2p
    inputf1_dt_k2p_w = inputf1, dt, k2p, w
    if w is None:
        w = 1
    p, _ = curve_fit(srtm_fun_k2p_w, inputf1_dt_k2p_w, tac * w, p0=(1, 0.5), bounds=(0, [3, 10]))
    theta_0 = p[0]
    theta_2 = p[1]
    theta_1 = theta_0 * (k2p-theta_2)
    theta = np.array([theta_0, theta_1, theta_2])
    r1, k2, bp = kp.srtm_theta2kp(theta)
    tacf = srtm_fun_k2p(inputf1_dt_k2p, theta_0, theta_2)
    kps = {'r1': r1, 'k2': k2, 'bp': bp, 'tacf': tacf}
    return kps


# # # # # # # # # # # # # # # # # # # # # # # # # # # #
# #  model for ref input
# # # # # # # # # # # # # # # # # # # # # # # # # # # #
def exp_1_fun_t(t, a0, a1, b1):
    cr = a0 + a1 * np.exp(-b1 * t)
    return cr


def exp_1_fun(ts_te_w, a0, a1, b1):
    ts, te, w = ts_te_w
    if w is None:
        w = 1
    cr_dt_fun = a0 * (te-ts) - (a1 * (np.exp(-b1 * te) - np.exp(-b1 * ts))) / b1
    return cr_dt_fun * w


def exp_1(tac, dt, idx, w, fig):
    tac[tac < 0] = 0.0
    if w is None:
        w = np.ones_like(tac)
    ts_te_w = (dt[0, idx], dt[1, idx], w[idx])
    p0 = (1000, 5000, 10)
    p, _ = curve_fit(exp_1_fun, ts_te_w, tac[idx] * w[idx], p0)
    a0, a1, b1 = p
    t1 = np.arange(np.amax(dt))
    tac1f = exp_1_fun_t(t1, a0, a1, b1)
    if fig:
        print(p)
        mft = kt.dt2mft(dt)
        plt.plot(t1, tac1f, 'b', mft, tac, 'go')
        plt.show()
    return tac1f, p


def exp_2_fun_t(t, a0, a1, a2, b1, b2):
    cr = a0 + a1 * np.exp(-b1 * t) + a2 * np.exp(-b2 * t)
    return cr


def exp_2_fun(ts_te_w, a0, a1, a2, b1, b2):
    ts, te, w = ts_te_w
    if w is None:
        w = 1
    cr_dt_fun = a0 * (te-ts) - (a1 * (np.exp(-b1 * te) - np.exp(-b1 * ts))) / b1 - (
        a2 * (np.exp(-b2 * te) - np.exp(-b2 * ts))) / b2
    return cr_dt_fun * w


def exp_2(tac, dt, idx, w, fig):
    tac[tac < 0] = 0.0
    if w is None:
        w = np.ones_like(tac)
    ts_te_w = (dt[0, idx], dt[1, idx], w[idx])
    p0 = (1, 1, 1, 0, 0)
    p, _ = curve_fit(exp_2_fun, ts_te_w, tac[idx] * w[idx], p0)
    a0, a1, a2, b1, b2 = p
    t1 = np.arange(np.amax(dt))
    tac1f = exp_2_fun_t(t1, a0, a1, a2, b1, b2)
    if fig:
        print(p)
        mft = kt.dt2mft(dt)
        plt.plot(t1, tac1f, 'b', mft, tac, 'go')
        plt.show()
    return tac1f, p


def exp_am(tac, dt, idx, fig):
    tac[tac < 0] = 0.0
    mft = kt.dt2mft(dt)
    p0 = (0.1, 1, 0.1)
    # p, _ = curve_fit(exp_1_fun_t, mft[idx], tac[idx], p0=p0, bounds=(0.00000001, 2500))
    p, _ = curve_fit(exp_1_fun_t, mft[idx], tac[idx], p0=p0)
    a0, a1, b1 = p
    t1 = np.arange(np.amax(dt))
    tac1f = exp_1_fun_t(t1, a0, a1, b1)
    if fig:
        print(p)
        plt.plot(t1, tac1f, 'b', mft, tac, 'go')
        plt.show()
    return tac1f, p


def feng_srtm_fun(ts_te_w, a0, a1, a2, a3, b0, b1, b2, b3):
    ts, te, w = ts_te_w
    if w is None:
        w = 1
    cr_dt_fun = (a0 * a3 *
                 (b0**2 * te / (b0**4 * np.exp(b0 * te) - 2 * b0**3 * b3 * np.exp(b0 * te) +
                                b0**2 * b3**2 * np.exp(b0 * te)) - b0 * b3 * te /
                  (b0**4 * np.exp(b0 * te) - 2 * b0**3 * b3 * np.exp(b0 * te) +
                   b0**2 * b3**2 * np.exp(b0 * te)) + 2 * b0 /
                  (b0**4 * np.exp(b0 * te) - 2 * b0**3 * b3 * np.exp(b0 * te) +
                   b0**2 * b3**2 * np.exp(b0 * te)) - b3 /
                  (b0**4 * np.exp(b0 * te) - 2 * b0**3 * b3 * np.exp(b0 * te) +
                   b0**2 * b3**2 * np.exp(b0 * te))) - a0 * a3 *
                 (b0**2 * ts / (b0**4 * np.exp(b0 * ts) - 2 * b0**3 * b3 * np.exp(b0 * ts) +
                                b0**2 * b3**2 * np.exp(b0 * ts)) - b0 * b3 * ts /
                  (b0**4 * np.exp(b0 * ts) - 2 * b0**3 * b3 * np.exp(b0 * ts) +
                   b0**2 * b3**2 * np.exp(b0 * ts)) + 2 * b0 /
                  (b0**4 * np.exp(b0 * ts) - 2 * b0**3 * b3 * np.exp(b0 * ts) +
                   b0**2 * b3**2 * np.exp(b0 * ts)) - b3 /
                  (b0**4 * np.exp(b0 * ts) - 2 * b0**3 * b3 * np.exp(b0 * ts) +
                   b0**2 * b3**2 * np.exp(b0 * ts))) + a0 * a3 * np.exp(-b3 * ts) /
                 (b3 * (b0**2 - 2*b0*b3 + b3**2)) - a0 * a3 * np.exp(-b3 * te) /
                 (b3 * (b0**2 - 2*b0*b3 + b3**2)) - a1 * a3 * np.exp(-b1 * ts) /
                 (b1**2 - b1*b3) + a1 * a3 * np.exp(-b1 * te) /
                 (b1**2 - b1*b3) + a1 * a3 * np.exp(-b0 * ts) /
                 (b0**2 - b0*b3) - a1 * a3 * np.exp(-b0 * te) /
                 (b0**2 - b0*b3) + a1 * a3 * np.exp(-b3 * ts) /
                 (b3 * (b1-b3)) - a1 * a3 * np.exp(-b3 * te) /
                 (b3 * (b1-b3)) - a1 * a3 * np.exp(-b3 * ts) /
                 (b3 *
                  (b0-b3)) + a1 * a3 * np.exp(-b3 * te) / (b3 *
                                                           (b0-b3)) - a2 * a3 * np.exp(-b2 * ts) /
                 (b2**2 - b2*b3) + a2 * a3 * np.exp(-b2 * te) /
                 (b2**2 - b2*b3) + a2 * a3 * np.exp(-b0 * ts) /
                 (b0**2 - b0*b3) - a2 * a3 * np.exp(-b0 * te) /
                 (b0**2 - b0*b3) + a2 * a3 * np.exp(-b3 * ts) /
                 (b3 *
                  (b2-b3)) - a2 * a3 * np.exp(-b3 * te) / (b3 *
                                                           (b2-b3)) - a2 * a3 * np.exp(-b3 * ts) /
                 (b3 * (b0-b3)) + a2 * a3 * np.exp(-b3 * te) / (b3 * (b0-b3))) / (te-ts)

    return cr_dt_fun * w


def feng_fun_t(t, a0, a1, a2, b0, b1, b2):
    cp = a0 * t * np.exp(-b0 * t) + a1 * np.exp(-b1 * t) - a1 * np.exp(-b0 * t) + a2 * np.exp(
        -b2 * t) - a2 * np.exp(-b0 * t)
    return cp


def feng_srtm_fun_t(t, a0, a1, a2, a3, b0, b1, b2, b3):
    cr1 = a0 * a3 * (
        -b0 * t * np.exp(b3 * t) /
        (b0**2 * np.exp(b0 * t) - 2 * b0 * b3 * np.exp(b0 * t) + b3**2 * np.exp(b0 * t)) +
        b3 * t * np.exp(b3 * t) /
        (b0**2 * np.exp(b0 * t) - 2 * b0 * b3 * np.exp(b0 * t) + b3**2 * np.exp(b0 * t)) -
        np.exp(b3 * t) /
        (b0**2 * np.exp(b0 * t) - 2 * b0 * b3 * np.exp(b0 * t) + b3**2 * np.exp(b0 * t))) * np.exp(
            -b3 * t)

    cr2 = a0 * a3 * np.exp(-b3 * t) / (b0**2 - 2*b0*b3 + b3**2)
    cr3 = -a1 * a3 / (b1 * np.exp(b1 * t) - b3 * np.exp(b1 * t))
    cr4 = a1 * a3 / (b0 * np.exp(b0 * t) - b3 * np.exp(b0 * t))
    cr5 = a1 * a3 * np.exp(-b3 * t) / (b1-b3)
    cr6 = -a1 * a3 * np.exp(-b3 * t) / (b0-b3)
    cr7 = -a2 * a3 / (b2 * np.exp(b2 * t) - b3 * np.exp(b2 * t))
    cr8 = a2 * a3 / (b0 * np.exp(b0 * t) - b3 * np.exp(b0 * t))
    cr9 = a2 * a3 * np.exp(-b3 * t) / (b2-b3)
    cr10 = -a2 * a3 * np.exp(-b3 * t) / (b0-b3)
    cr = np.stack((cr1, cr2, cr3, cr4, cr5, cr6, cr7, cr8, cr9, cr10))
    cr[np.isinf(cr)] = 0
    cr[np.isnan(cr)] = 0
    cr = np.sum(cr, axis=0)
    return cr


def feng_srtm(tac, dt, w, fig):
    tac[tac < 0] = 0.0
    if w is None:
        w = 1
    ts_te_w = (dt[0,], dt[1,], w)
    # p0 = [3.90671734e+00, 4.34910151e+02, 9.22189828e+01, 1.35949657e-02,
    #       4.56109635e-02, 4.53841116e-02, 4.54180443e-02, 7.71163349e-04]
    p0 = [1, 2, 3, 4, 0.1, 0.2, 0.3, 0.4]
    # p0 = [6.85579165e-05, -2.08643110e+01, -1.81889002e+02, 7.16906660e+00,
    #       4.21217390e-04, 7.23514957e-02, 7.84986975e-02, 8.27340347e-02]
    p, _ = curve_fit(feng_srtm_fun, ts_te_w, tac * w, p0=p0)
    a0, a1, a2, a3, b0, b1, b2, b3 = p
    t1 = np.arange(np.amax(dt))
    tac1f = feng_srtm_fun_t(t1, a0, a1, a2, a3, b0, b1, b2, b3)
    print(tac1f)
    if fig:
        print(p)
        cp1f = feng_fun_t(t1, a0, a1, a2, b0, b1, b2)
        mft = kt.dt2mft(dt)
        plt.plot(t1, cp1f, 'r', t1, tac1f, 'b', mft, tac, 'go')
        plt.show()
    return tac1f, p



__author__ = 'jieqing jiao'
__contact__ = 'jieqing.jiao@gmail.com'

import numpy as np
from tqdm import trange

from ..kt import dt2tdur
from ..tac import TAC
from .regions import extract_regional_values


def image_to_parametric(pet_image, dt, model_name, model_inputs, km_outputs, thr=0.005):
    parametric_images = []
    tac = TAC(pet_image[0, 0, 0,] * 0 + 1, dt)
    tac.run_model(model_name, model_inputs)
    km_outputs = list({c.lower() for c in km_outputs} & set(tac.km_results.keys()))
    for _ in range(len(km_outputs)):
        parametric_images.append(np.zeros(pet_image.shape[0:3]))
    pet_image_fit = np.zeros(pet_image.shape)
    thr = thr * np.amax(pet_image)
    mask = np.argwhere(np.mean(pet_image, axis=-1) > thr)
    for i in trange(mask.shape[0]):
        tac = TAC(pet_image[mask[i][0], mask[i][1], mask[i][2],], dt)
        tac.run_model(model_name, model_inputs)
        # # #
        # plt.plot(tac.mft, tac.km_results['tacf'],'r', tac.mft, tac.tac, 'go')
        # plt.show()
        # # #
        if 'tacf' in tac.km_results:
            pet_image_fit[mask[i][0], mask[i][1], mask[i][2],] = tac.km_results['tacf']
        # # #
        # print(tac.km_results)
        # # #
        for p in range(len(km_outputs)):
            parametric_images[p][mask[i][0], mask[i][1],
                                 mask[i][2],] = tac.km_results[km_outputs[p].lower()]
    parametric_images_dict = dict(zip(km_outputs, parametric_images))
    if 'tacf' in tac.km_results:
        return parametric_images_dict, pet_image_fit
    return parametric_images_dict


def parametric_to_image(parametric_images_dict, dt, model, km_inputs):
    parametric_images = list(parametric_images_dict.values())
    pet_image = np.zeros(parametric_images[0].shape[0:3] + (dt.shape[-1],))
    mask = np.argwhere(parametric_images[0] != 0)
    for i in range(mask.shape[0]):
        km_inputs_local = km_inputs.copy()
        for p in parametric_images_dict.keys():
            km_inputs_local.update({
                p.lower(): parametric_images_dict[p][mask[i][0], mask[i][1], mask[i][2]]})
        tac = TAC([], dt)
        getattr(tac, 'run_' + model + '_para2tac')(**km_inputs_local)
        pet_image[mask[i][0], mask[i][1], mask[i][2],] = tac.km_results['tacf']
    return pet_image


def image_to_suvr_with_parcellation(pet_image, dt, parcellation, reference_region_labels,
                                    selected_frame_index):
    reference_regional_tac = extract_regional_values(pet_image, parcellation,
                                                     reference_region_labels)
    image_suvr = image_to_suvr_with_reference_tac(pet_image, dt, reference_regional_tac,
                                                  selected_frame_index)
    return image_suvr


def image_to_suvr_with_reference_tac(pet_image, dt, reference_regional_tac, selected_frame_index):
    tdur = dt2tdur(dt)
    image = np.multiply(pet_image[:, :, :, selected_frame_index], tdur[selected_frame_index])
    ref = np.multiply(reference_regional_tac[selected_frame_index], tdur[selected_frame_index])
    image_suvr = np.sum(image, axis=-1) / np.sum(ref)
    return image_suvr



"""Clean API"""
import logging
from pathlib import Path

from . import readers

log = logging.getLogger(__name__)


def kinetic_model(src, dst=None, params=None, model='srtmb_basis', input_interp_method='linear',
                  w=None, r1=1, k2p=0.000250, beta_lim=None, n_beta=40, linear_phase_start=500,
                  linear_phase_end=None, km_outputs=None, thr=0.1, fig=False):
    """
    Args:
      src (Path or str): input patient directory or filename
      dst (Path or str): output directory (default: `src` directory)
      params (Path or str): config (relative to `src` directory)
      model (str): any model from `niftypad.models` (see `niftypad.models.NAMES`)
      input_interp_method (str): the interpolation method for getting reference input:
        linear, cubic, exp_1, exp_2, feng_srtm
      w (ndarray): weights for weighted model fitting
      r1 (float): a pre-chosen value between 0 and 1 for r1, used in srtmb_asl_basis
      k2p (float): a pre-chosen value for k2p, in second^-1, used in
        srtmb_k2p_basis, logan_ref_k2p, mrtm_k2p
      beta_lim (list[int]): [beta_min, beta_max] for setting the lower and upper limits
        of beta values in basis functions, used in srtmb_basis, srtmb_k2p_basis, srtmb_asl_basis
      n_beta (int): number of beta values/basis functions, used in
        srtmb_basis, srtmb_k2p_basis, srtmb_asl_basis
      linear_phase_start (int): start time of linear phase in seconds, used in logan_ref,
        logan_ref_k2p, mrtm, mrtm_k2p
      linear_phase_end (int): end time of linear phase in seconds, used in logan_ref,
        logan_ref_k2p, mrtm, mrtm_k2p
      km_outputs (list[str]): the kinetic parameters to save, e.g. ['R1', 'k2', 'BP']
      thr (float): threshold value between 0 and 1. Used to mask out voxels with mean value
        over time exceeding `thr * max(image value)`
      fig (bool): whether to show a figure to check model fitting
    """
    import nibabel as nib
    import numpy as np

    from niftypad import basis
    from niftypad.image_process.parametric_image import image_to_parametric
    from niftypad.models import get_model_inputs
    from niftypad.tac import Ref

    src_path = Path(src)
    if src_path.is_dir():
        fpath = next(src_path.glob('*.nii'))
    else:
        fpath = src_path
        src_path = fpath.parent
    log.debug("file:%s", fpath)

    if dst is None:
        dst_path = src_path
    else:
        dst_path = Path(dst)
        assert dst_path.is_dir()

    meta = readers.find_meta(src_path, filter(None, [params, fpath.stem]))
    dt = np.asarray(meta['dt'])
    ref = np.asarray(meta['ref'])
    ref = Ref(ref, dt)
    # change ref interpolation to selected method
    ref.run_interp(input_interp_method=input_interp_method)

    log.debug("looking for first `*.nii` file in %s", src_path)
    img = nib.load(fpath)
    # pet_image = img.get_fdata(dtype=np.float32)
    pet_image = np.asanyarray(img.dataobj)

    # basis functions
    if beta_lim is None:
        beta_lim = [0.01 / 60, 0.3 / 60]
    # change ref.inputf1cubic -> ref.input_interp_1
    b = basis.make_basis(ref.input_interp_1, dt, beta_lim=beta_lim, n_beta=n_beta, w=w, k2p=k2p)

    if km_outputs is None:
        km_outputs = ['R1', 'k2', 'BP']
    # change ref.inputf1cubic -> ref.input_interp_1
    user_inputs = {
        'dt': dt, 'ref': ref, 'inputf1': ref.input_interp_1, 'w': w, 'r1': r1, 'k2p': k2p,
        'beta_lim': beta_lim, 'n_beta': n_beta, 'b': b, 'linear_phase_start': linear_phase_start,
        'linear_phase_end': linear_phase_end, 'fig': fig}
    model_inputs = get_model_inputs(user_inputs, model)
    # log.debug("model_inputs:%s", model_inputs)

    parametric_images_dict, pet_image_fit = image_to_parametric(pet_image, dt, model, model_inputs,
                                                                km_outputs, thr=thr)
    for kp in parametric_images_dict:
        nib.save(nib.Nifti1Image(parametric_images_dict[kp], img.affine),
                 f"{dst_path / fpath.stem}_{model}_{kp}_{fpath.suffix}")
    nib.save(nib.Nifti1Image(pet_image_fit, img.affine),
             f"{dst_path / fpath.stem}_{model}_fit_{fpath.suffix}")



__author__ = 'jieqing jiao'

import matplotlib.pyplot as plt
import numpy as np

from . import kt


def make_basis(inputf1, dt, beta_lim=None, n_beta=128, beta_space='log', w=None, k2p=None,
               fig=False):
    if beta_lim is None:
        beta_lim = [10e-6, 10e-1]

    # dt can skip frames if needed

    # calculate input based on dt
    tdur = kt.dt2tdur(dt)
    input = kt.int2dt(inputf1, dt)
    t1 = np.arange(np.amax(dt))

    # generate basis functions and m based on dt
    if beta_space == 'log':
        beta = np.logspace(start=np.log10(beta_lim[0]), stop=np.log10(beta_lim[1]), num=n_beta,
                           endpoint=False)
    elif beta_space == 'natural':
        beta = np.zeros(n_beta)
        for i in range(0, n_beta):
            beta[i] = -1 / t1[-1] * np.log(
                (np.exp(-beta_lim[0] * t1[-1]) - np.exp(-beta_lim[1] * t1[-1])) / n_beta * i +
                np.exp(-beta_lim[1] * t1[-1]))
    basis = np.zeros((beta.size, tdur.size))
    basis_w = np.zeros((beta.size, tdur.size))
    m = np.zeros((beta.size * 2, tdur.size))
    m_w = np.zeros((beta.size * 2, tdur.size))
    basis_k2p = np.zeros((beta.size, tdur.size)) # no need to calculate basis_k2p_w

    for i in range(0, beta.size):
        # make basis
        # inputf1[inputf1 < 0] = 0.0
        basis1 = np.convolve(inputf1, np.exp(-beta[i] * t1))
        # basis1 = scipy.signal.convolve(inputf1, np.exp(-beta[i]*t1), method='direct')
        basis[i, :] = kt.int2dt(basis1, dt)
        if w is not None:
            basis_w[i, :] = basis[i, :] * w
        # make m
        a = np.column_stack((input, basis[i, :]))
        q, r = np.linalg.qr(a)
        m[i * 2:i*2 + 2, :] = np.linalg.solve(r, q.T)
        if w is not None:
            a = np.dot(np.diag(w), a)
            q, r = np.linalg.qr(a)
            m_w[i * 2:i*2 + 2, :] = np.linalg.solve(r, q.T)
        # make basis_k2p with fixed k2p
        if k2p is not None:
            basis_k2p[i, :] = input + (k2p - beta[i]) * basis[i, :]

    if w is None:
        basis_w = basis
        m_w = m

    basis = {
        'beta': beta, 'basis': basis, 'basis_w': basis_w, 'm': m, 'm_w': m_w,
        'basis_k2p': basis_k2p, 'input': input, 'w': w, 'k2p': k2p, 'dt': dt}

    if fig:
        plt.plot(kt.dt2mft(dt), basis['basis'].T)
        plt.show()
    return basis



__author__ = 'jieqing jiao'
__email__ = "jieqing.jiao@gmail.com"

import numpy as np


def srtm_theta2kp(theta):
    # theta - output from models/srtmb
    # theta[0] - linear coeff of input(Cr)
    # theta[1] - linear coeff of the conv
    # theta[2] - exp coeff (no -)

    r1 = theta[0]                       # R1
    k2 = theta[1] + theta[0] * theta[2] # k2
    bp = k2 / theta[2] - 1              # BPnd
    return r1, k2, bp


def srtm_kp2theta(r1, k2, bp):
    theta = np.zeros(3)
    theta[0] = r1
    theta[1] = k2 * (1 - r1 / (1+bp))
    theta[2] = k2 / (1+bp)
    return theta



from os import getenv
from pathlib import Path

from pytest import fixture, skip

HOME = Path(getenv("DATA_ROOT", "~")).expanduser()


@fixture(scope="session")
def folder_in():
    AMYPAD = HOME / "AMYPAD"
    if not AMYPAD.is_dir():
        skip(f"Cannot find AMYPAD in ${{DATA_ROOT:-~}} ({HOME})")
    return AMYPAD



__author__ = 'jieqing jiao'
__email__ = "jieqing.jiao@gmail.com"

from math import cos, sin

import numpy as np
import numpy.matlib
from numpy.linalg import solve
from scipy.interpolate import interpn


def rigid_p2matrix(translation, rotation):
    '''
    :param translation: shape (3,)
    :param rotation: shape (3,)
    :return T: shape(4,4) translation matrix
    :return R: shape(4,4) rotation matrix
    '''
    t_matrix = np.eye(4)
    r_x_matrix = np.eye(4)
    r_y_matrix = np.eye(4)
    r_z_matrix = np.eye(4)
    t_matrix[:3, -1] = translation
    phi = rotation[0]
    r_x_matrix[1, 1] = cos(phi)
    r_x_matrix[1, 2] = sin(phi)
    r_x_matrix[2, 1] = -sin(phi)
    r_x_matrix[2, 2] = cos(phi)
    phi = rotation[1]
    r_y_matrix[0, 0] = cos(phi)
    r_y_matrix[0, 2] = sin(phi)
    r_y_matrix[2, 0] = -sin(phi)
    r_y_matrix[2, 2] = cos(phi)
    phi = rotation[2]
    r_z_matrix[0, 0] = cos(phi)
    r_z_matrix[0, 1] = sin(phi)
    r_z_matrix[1, 0] = -sin(phi)
    r_z_matrix[1, 1] = cos(phi)
    r_matrix = r_x_matrix @ r_y_matrix @ r_z_matrix
    return t_matrix, r_matrix


def affine_transform_centre(xv, yv, zv, translation, rotation, centre, order='T*Tc_i*R*Tc'):
    '''
    :param xv: mesh grid
    :param yv: mesh grid
    :param zv: mesh grid
    :param translation:
    :param rotation:
    :param centre: centre for rotation
    :param order: 'Tc_i*R*Tc*T' or 'T*Tc_i*R*Tc'
    :return:
    '''
    t_matrix, r_matrix = rigid_p2matrix(translation, rotation)
    t_c_i_matrix, _ = rigid_p2matrix(centre, [0, 0, 0])
    t_c_matrix, _ = rigid_p2matrix([-c for c in centre], [0, 0, 0])
    m = np.eye(4)
    if order == 'T*Tc_i*R*Tc':
        m = t_matrix @ t_c_i_matrix @ r_matrix @ t_c_matrix
    if order == 'Tc_i*R*Tc*T':
        m = t_c_i_matrix @ r_matrix @ t_c_matrix @ t_matrix

    coordinates = np.row_stack((xv.flatten(), yv.flatten(), zv.flatten(), np.ones(xv.size)))
    coordinates_new = m @ coordinates
    xv_new = coordinates_new[0, :].reshape(xv.shape)
    yv_new = coordinates_new[1, :].reshape(yv.shape)
    zv_new = coordinates_new[2, :].reshape(zv.shape)
    return xv_new, yv_new, zv_new


def d_coordinates_over_d_transform(xv, yv, zv, rotation, centre):
    xv = xv.flatten()
    yv = yv.flatten()
    zv = zv.flatten()
    p4, p5, p6 = rotation[:3]
    c1, c2, c3 = centre[:3]
    ones = np.ones(xv.size)

    dx_over_dp = np.column_stack(
        (ones * 1, ones * 0, ones * 0, ones * 0,
         cos(p5) * zv - c3 * cos(p5) - yv * sin(p5) * sin(p6) + c1 * cos(p6) * sin(p5) +
         c2 * sin(p5) * sin(p6) - cos(p6) * xv * sin(p5), c1 * cos(p5) * sin(p6) -
         c2 * cos(p5) * cos(p6) + cos(p5) * cos(p6) * yv - cos(p5) * xv * sin(p6)))

    dy_over_dp = np.column_stack(
        (ones * 0, ones * 1, ones * 0, c2 * (cos(p6) * sin(p4) + cos(p4) * sin(p5) * sin(p6)) -
         c1 * (sin(p4) * sin(p6) - cos(p4) * cos(p6) * sin(p5)) + xv *
         (sin(p4) * sin(p6) - cos(p4) * cos(p6) * sin(p5)) - yv *
         (cos(p6) * sin(p4) + cos(p4) * sin(p5) * sin(p6)) - c3 * cos(p4) * cos(p5) +
         cos(p4) * cos(p5) * zv, c3 * sin(p4) * sin(p5) - zv * sin(p4) * sin(p5) +
         c1 * cos(p5) * cos(p6) * sin(p4) + c2 * cos(p5) * sin(p4) * sin(p6) -
         cos(p5) * cos(p6) * xv * sin(p4) - cos(p5) * yv * sin(p4) * sin(p6),
         c1 * (cos(p4) * cos(p6) - sin(p4) * sin(p5) * sin(p6)) + c2 *
         (cos(p4) * sin(p6) + cos(p6) * sin(p4) * sin(p5)) - xv *
         (cos(p4) * cos(p6) - sin(p4) * sin(p5) * sin(p6)) - yv *
         (cos(p4) * sin(p6) + cos(p6) * sin(p4) * sin(p5))))

    dz_over_dp = np.column_stack(
        (ones * 0, ones * 0, ones * 1, c2 * (cos(p4) * cos(p6) - sin(p4) * sin(p5) * sin(p6)) -
         c1 * (cos(p4) * sin(p6) + cos(p6) * sin(p4) * sin(p5)) + xv *
         (cos(p4) * sin(p6) + cos(p6) * sin(p4) * sin(p5)) - yv *
         (cos(p4) * cos(p6) - sin(p4) * sin(p5) * sin(p6)) + c3 * cos(p5) * sin(p4) -
         cos(p5) * zv * sin(p4), c3 * cos(p4) * sin(p5) - cos(p4) * zv * sin(p5) +
         c1 * cos(p4) * cos(p5) * cos(p6) + c2 * cos(p4) * cos(p5) * sin(p6) -
         cos(p4) * cos(p5) * cos(p6) * xv - cos(p4) * cos(p5) * yv * sin(p6),
         xv * (cos(p6) * sin(p4) + cos(p4) * sin(p5) * sin(p6)) - c2 *
         (sin(p4) * sin(p6) - cos(p4) * cos(p6) * sin(p5)) - c1 *
         (cos(p6) * sin(p4) + cos(p4) * sin(p5) * sin(p6)) + yv *
         (sin(p4) * sin(p6) - cos(p4) * cos(p6) * sin(p5))))

    return dx_over_dp, dy_over_dp, dz_over_dp


def image_rigid_transform_3d(image, translation, rotation):
    x_range = np.arange(image.shape[0])
    y_range = np.arange(image.shape[1])
    z_range = np.arange(image.shape[2])
    xv, yv, zv = np.meshgrid(x_range, y_range, z_range, indexing='ij')
    centre = [xv.mean(), yv.mean(), zv.mean()]
    xv_new, yv_new, zv_new = affine_transform_centre(xv, yv, zv, translation, rotation, centre)
    image_new = interpn((x_range, y_range, z_range), image,
                        np.vstack(
                            (xv_new.flatten(), yv_new.flatten(), zv_new.flatten())).transpose(),
                        bounds_error=False)
    image_new = image_new.reshape(image.shape)
    return image_new


def image_registration_3d_gn(image, image_ref, translation, rotation, n_iter=100):
    '''
    rigid registration of two 3d images using Gauss-Newton method
    :param image:
    :param image_ref:
    :param translation:
    :param rotation:
    :param n_iter:
    :return:
    '''
    x_range = np.arange(image.shape[0])
    y_range = np.arange(image.shape[1])
    z_range = np.arange(image.shape[2])
    xv, yv, zv = np.meshgrid(x_range, y_range, z_range, indexing='ij')
    centre = [xv.mean(), yv.mean(), zv.mean()]
    for i in range(n_iter):
        print(i)
        xv_new, yv_new, zv_new = affine_transform_centre(xv, yv, zv, translation, rotation, centre)
        image_new = interpn(
            (x_range, y_range, z_range), image,
            np.vstack((xv_new.flatten(), yv_new.flatten(), zv_new.flatten())).transpose(),
            bounds_error=False)
        image_new = image_new.reshape(image.shape)
        dx_over_dp, dy_over_dp, dz_over_dp = d_coordinates_over_d_transform(
            xv, yv, zv, rotation, centre)
        image_new_g = image_new * 1
        image_new_g[np.isnan(image_new)] = 0
        image_gradient = np.gradient(image_new_g)
        image_gradient_x = np.matlib.repmat(image_gradient[0].flatten(), dx_over_dp.shape[-1],
                                            1).transpose()
        image_gradient_y = np.matlib.repmat(image_gradient[1].flatten(), dy_over_dp.shape[-1],
                                            1).transpose()
        image_gradient_z = np.matlib.repmat(image_gradient[2].flatten(), dz_over_dp.shape[-1],
                                            1).transpose()
        jacobian = (image_gradient_x*dx_over_dp + image_gradient_y*dy_over_dp +
                    image_gradient_z*dz_over_dp)
        image_diff = image_new - image_ref
        image_diff[np.isnan(image_diff)] = 0
        a = jacobian.transpose() @ jacobian
        update = solve(a, jacobian.transpose()) @ image_diff.flatten()
        translation = np.subtract(translation, update[:3])
        rotation = np.subtract(rotation, update[3:])
        print(translation)
        print(rotation)
        if all(abs(update) < 0.01):
            break
    return translation, rotation, image_new_g


def image_registration_3d_plus_t(image_t, image_ref_t, translation_t, rotation_t,
                                 registration_index, n_iter=100):
    image_new_t = image_t * 1
    for t in registration_index:
        translation_t[t], rotation_t[t], image_new_t[:, :, :, t] = \
            image_registration_3d_gn(
                image_t[:, :, :, t], image_ref_t[:, :, :, t],
                translation=translation_t[t], rotation=rotation_t[t], n_iter=n_iter)
    return translation_t, rotation_t, image_new_t



__author__ = 'jieqing jiao'
__email__ = "jieqing.jiao@gmail.com"

import nibabel as nib
import numpy as np


def extract_regional_values(image, parcellation, labels):
    idx = labels_to_index(parcellation, labels)
    regional_values = np.mean(image[idx,], axis=0)
    return regional_values


def labels_to_index(parcellation, labels):
    parcellation = np.squeeze(parcellation)
    idx = np.zeros(parcellation.shape, dtype='bool')
    for i in range(len(labels)):
        idx = np.logical_or(idx, parcellation == labels[i])
    return idx


def extract_regional_values_image_file(image_file, parcellation_file):
    image = nib.load(image_file)
    image_data = image.get_data()
    n_frames = 1
    if image_data.ndim == 4:
        n_frames = image_data.shape[-1]
    parcellation_img = nib.load(parcellation_file)
    parcellation = parcellation_img.get_data()
    regions_label = np.unique(parcellation)
    regions_data = np.zeros((regions_label.size, n_frames))
    regions_data = np.squeeze(regions_data)
    for i in range(regions_label.size):
        regions_data[i] = extract_regional_values(image_data, parcellation, [regions_label[i]])
    return regions_data, regions_label



__author__ = 'jieqing jiao'
__email__ = "jieqing.jiao@gmail.com"

from . import kt, models


class TAC:
    def __init__(self, tac, dt):
        self.tac = tac
        self.dt = dt
        self.mft = kt.dt2mft(dt)

        self.km_results = []

    def run_model(self, model_name, model_inputs):
        km_results = {'model_name': model_name}
        model = getattr(models, model_name)
        kps = model(self.tac, **model_inputs)
        km_results.update(kps)
        self.km_results = km_results

    def run_model_para2tac(self, model_name, model_inputs):
        km_results = {'model_name': model_name}
        model = getattr(models, model_name)
        kps = model(**model_inputs)
        km_results.update(kps)
        self.km_results = km_results


class Ref:
    def __init__(self, tac, dt):
        self.tac = tac
        self.dt = dt
        self.inputf1_feng_srtm = []
        self.inputf1 = []
        self.inputf1cubic = []
        self.inputf1_exp1 = []
        self.inputf1_exp2 = []
        self.inputf1_exp_am = []
        self.input_interp_method = []
        self.input_interp_1 = []

    def run_feng_srtm(self, w=None):
        self.inputf1_feng_srtm, _ = models.feng_srtm(self.tac, self.dt, w=w, fig=True)
        self.input_interp_1 = self.inputf1_feng_srtm

    def interp_1(self):
        mft = kt.dt2mft(self.dt)
        self.inputf1 = kt.interpt1(mft, self.tac, self.dt)
        self.input_interp_1 = self.inputf1

    def interp_1cubic(self):
        mft = kt.dt2mft(self.dt)
        self.inputf1cubic = kt.interpt1cubic(mft, self.tac, self.dt)
        self.input_interp_1 = self.inputf1cubic

    def run_exp1(self, w=None, idx_to_fit=None, fill_in_seconds=None):
        self.interp_1()
        self.inputf1_exp1 = self.inputf1
        inputf1_exp1 = 0
        if idx_to_fit is not None:
            inputf1_exp1, _ = models.exp_1(self.tac, self.dt, idx=idx_to_fit, w=w, fig=True)
        if fill_in_seconds is not None:
            self.inputf1_exp1[fill_in_seconds] = inputf1_exp1[fill_in_seconds]
        self.input_interp_1 = self.inputf1_exp1

    def run_exp2(self, w=None, idx_to_fit=None, fill_in_seconds=None):
        self.interp_1()
        self.inputf1_exp2 = self.inputf1
        inputf1_exp2 = 0
        if idx_to_fit is not None:
            inputf1_exp2, _ = models.exp_2(self.tac, self.dt, idx=idx_to_fit, w=w, fig=True)
        if fill_in_seconds is not None:
            self.inputf1_exp2[fill_in_seconds] = inputf1_exp2[fill_in_seconds]
        self.input_interp_1 = self.inputf1_exp2

    def run_exp_am(self, idx_to_fit=None, fill_in_seconds=None):
        self.interp_1()
        self.inputf1_exp_am = self.inputf1
        inputf1_exp_am = 0
        if idx_to_fit is not None:
            inputf1_exp_am, _ = models.exp_am(self.tac, self.dt, idx=idx_to_fit, fig=True)
        if fill_in_seconds is not None:
            self.inputf1_exp_am[fill_in_seconds] = inputf1_exp_am[fill_in_seconds]
        self.input_interp_1 = self.inputf1_exp_am

    def run_interp(self, input_interp_method='linear', **kwargs):
        input_interp_methods = {
            'linear': 'interp_1', 'cubic': 'interp_1cubic', 'exp_1': 'run_exp1',
            'exp_2': 'run_exp2', 'exp_am': 'run_exp_am', 'feng_srtm': 'run_feng_srtm'}
        self.input_interp_method = input_interp_method
        interp_method = getattr(self,
                                input_interp_methods.get(input_interp_method, input_interp_method))
        return interp_method(**kwargs)



__author__ = 'jieqing jiao'
__email__ = "jieqing.jiao@gmail.com"

import matplotlib.pyplot as plt
import numpy as np
from scipy.interpolate import interp1d


def dt2mft(dt):
    mft = np.mean(dt, axis=0)
    return mft


def dt2tdur(dt):
    tdur = dt[1,] - dt[0,]
    return tdur


def mft2tdur(mft):
    print('mind the gap')
    tdur = np.zeros_like(mft)
    tdur[0] = mft[0] * 2
    for i in range(1, len(mft)):
        tdur[i] = (mft[i] - np.sum(tdur)) * 2
    return tdur


def tdur2dt(tdur):
    print('mind the gap')
    st = np.zeros_like(tdur)
    et = np.zeros_like(tdur)
    st[0] = 0
    et[0] = tdur[0]
    for i in range(1, len(tdur)):
        st[i] = et[i - 1]
        et[i] = st[i] + tdur[i]
    dt = [st, et]
    return dt


def mft2dt(mft):
    print('mind the gap')
    tdur = mft2tdur(mft)
    dt = tdur2dt(tdur)
    return dt


def tdur2mft(tdur):
    print('mind the gap')
    dt = tdur2dt(tdur)
    mft = dt2mft(dt)
    return mft


def dt_has_gaps(dt):
    dt_has_gaps = False
    unique, counts = np.unique(dt, return_counts=True)
    gap_ends = unique[counts == 1]
    if len(gap_ends) > 2:
        dt_has_gaps = True
    return dt_has_gaps


def dt_fill_gaps(dt):
    unique, counts = np.unique(dt, return_counts=True)
    gap_ends = unique[counts == 1]
    if len(gap_ends) > 2:
        gap_ends = gap_ends[1:-1]
        gap_starts_index = range(0, len(gap_ends), 2)
        for i in gap_starts_index:
            # print(gap_ends[i:i+2])
            dt_gap = np.unique(np.floor(np.linspace(gap_ends[i], gap_ends[i + 1],
                                                    10))).astype('int16')
            tdur_to_insert = np.diff(dt_gap)
            dt_to_insert = tdur2dt(tdur_to_insert) + gap_ends[i]
            dt = np.insert(dt, np.where(dt[1,] == gap_ends[i])[0] + 1, dt_to_insert, axis=-1)
    return dt


def tac_dt_fill_coffee_break(tac, dt, fig=False):
    # interpolate tac with coffee break using cubic interpolation
    mft = dt2mft(dt)
    tac_inputf1cubic = interpt1cubic(mft, tac, dt)
    dt_no_gaps = dt_fill_gaps(dt)
    tac_no_gaps = int2dt(tac_inputf1cubic, dt_no_gaps)
    if fig:
        plt.plot(mft, tac, '.')
        plt.plot(dt2mft(dt_no_gaps), tac_no_gaps, 'r')
        plt.show()
    return tac_no_gaps, dt_no_gaps


# # # #


def int2dt(f1, dt):
    # in case dt is not in whole seconds
    dt = np.rint(dt).astype(int)
    tdur = dt2tdur(dt)
    f = np.zeros(tdur.size)
    for j in range(0, tdur.size):
        f[j] = np.sum(f1[dt[0, j]:dt[1, j]]) / tdur[j]
    return f


def interpt1(inputt, inputf, dt):
    # interpolate function
    t1 = np.arange(np.amax(dt))
    if inputt[0] > t1[0]:
        inputt = np.append(t1[0], inputt)
        inputf = np.append(0, inputf)
    inputff = interp1d(inputt, inputf, kind='linear', fill_value='extrapolate')
    inputf1 = inputff(t1)
    return inputf1


def interpt1cubic(inputt, inputf, dt):
    # interpolate function
    t1 = np.arange(np.amax(dt))
    if inputt[0] > t1[0]:
        inputt = np.append(t1[0], inputt)
        inputf = np.append(0, inputf)
    inputff = interp1d(inputt, inputf, kind='cubic', fill_value='extrapolate')
    inputf1 = inputff(t1)
    return inputf1



"""Metadata reader utilities"""
import errno
from pathlib import Path

CONFIG_NAMES = ['meta', 'conf', 'config', 'params', 'param', 'input']


def read_json(fd):
    from json import load
    return load(fd)


def read_yaml(fd):
    from oyaml import safe_load
    return safe_load(fd)


READERS = {'yml': read_yaml, 'yaml': read_yaml, 'json': read_json}


def read_meta(fpath):
    """Auto-detect `fpath`'s type and return a contents `dict`.

    Args:
      fpath (Path or str): filename
    Returns:
      dict: contents
    """
    fpath = Path(fpath)
    return READERS[fpath.suffix[1:].lower()](fpath.open())


def find_meta(src_path, config_name_hints=None):
    """Find and return metadata file contents

    Args:
      src_path (Path or str): where to look
      config_name_hints (list): stems to search for in `src_path`
    Returns:
      dict: contents
    """
    src_path = Path(src_path)
    if config_name_hints is None:
        config_names = CONFIG_NAMES
    else:
        config_names = list(config_name_hints) + CONFIG_NAMES
    for ext in sorted([''] + ['.' + i for i in READERS]):
        for meta in config_names:
            if (src_path / f'{meta}{ext}').is_file():
                return read_meta(src_path / f'{meta}{ext}')
    raise IOError(errno.ENOENT, "Input metadata not found")



__author__ = 'jieqing jiao'
__contact__ = 'jieqing.jiao@gmail.com'

import csv
import os

import nibabel as nib

from .parametric_image import image_to_parametric, parametric_to_image
from .transforms import image_registration_3d_plus_t


def kinetic_model_motion_correction(pet_image, dt, model, km_inputs_initial, km_inputs, km_outputs,
                                    initial_fit_index, motion_correction_index, n_iteration,
                                    translation_t, rotation_t):
    parametric_images_dict, _ = image_to_parametric(pet_image[:, :, :, initial_fit_index],
                                                    dt[:, initial_fit_index], model,
                                                    km_inputs_initial, km_outputs)
    pet_image_fit = parametric_to_image(parametric_images_dict, dt, model, km_inputs)
    for ii in range(n_iteration):
        print('iteration ' + str(ii + 1))
        print('updating motion ...')
        translation_t, rotation_t, pet_image_realigned = image_registration_3d_plus_t(
            pet_image, pet_image_fit, translation_t, rotation_t, motion_correction_index)
        print('updating kinetics ...')
        parametric_images_dict, pet_image_fit = image_to_parametric(pet_image_realigned, dt, model,
                                                                    km_inputs, km_outputs)
    return pet_image_realigned, translation_t, rotation_t, parametric_images_dict


def kinetic_model_motion_correction_file(pet_file, dt, model, km_inputs_initial, km_inputs,
                                         km_outputs, initial_fit_index, motion_correction_index,
                                         n_iteration, translation_t, rotation_t, save_file):
    pet_img = nib.load(pet_file)
    pet_img_data = pet_img.get_data()
    # pet_img_data[pet_img_data < 0] = 0
    pet_img_data[:, :, 0:1, :] = 0
    pet_image_realigned, translation_t, rotation_t, parametric_images_dict = \
        kinetic_model_motion_correction(
            pet_img_data, dt, model, km_inputs_initial, km_inputs, km_outputs, initial_fit_index,
            motion_correction_index, n_iteration, translation_t, rotation_t)
    if save_file:
        parametric_images_dict.update({'mc': pet_image_realigned})
        for kk in parametric_images_dict.keys():
            nib.save(nib.Nifti1Image(parametric_images_dict[kk].squeeze(), pet_img.affine),
                     os.path.splitext(pet_file)[0] + '_mc_' + model + '_' + kk + '.nii')
        motion_file_name = os.path.splitext(pet_file)[0] + '_' + model + '_motion' + '.csv'
        motion_file = open(motion_file_name, 'w')
        motion_writer = csv.writer(motion_file, dialect='excel-tab', delimiter=',')
        motion_writer.writerows(list(zip(translation_t, rotation_t)))
        motion_file.close()



__author__ = 'jieqing jiao'
__email__ = "jieqing.jiao@gmail.com"

# weighting scheme implementation based on https://www.ncbi.nlm.nih.gov/pubmed/16912378

import numpy as np

from . import kt


def get_weight(dt, isotope='18F', a=1, t=None, r=None, tac=None):

    if isotope == '18F':
        half_life = 20.39 * 60
    elif isotope == '12C':
        half_life = 109.77 * 60
    else:
        print('not known yet')

    decay_const = np.log(2) / half_life
    tdur = kt.dt2tdur(dt)
    dcf = decay_const * tdur / (np.exp(-decay_const * dt[0,]) - np.exp(-decay_const * dt[1,]))

    n_model = 5
    w_m = [None] * n_model
    i = 0
    w_m[i] = a * dcf * dcf * t / tdur / tdur
    i += 1
    w_m[i] = a * dcf * dcf * (t + 2*r) / tdur / tdur
    i += 1
    w_m[i] = a
    i += 1
    w_m[i] = a * dcf * dcf
    i += 1
    w_m[i] = a * tac * tdur * dcf

    for i in range(n_model):
        w_m[i] = 1 / w_m[i]

    return w_m


def get_weight_ppet(dt, tac, isotope='12C', a=1):
    if isotope == '18F':
        half_life = 20.39 * 60
    elif isotope == '12C':
        half_life = 109.77 * 60
    else:
        print('not known yet')
    decay_const = np.log(2) / half_life

    tdur = kt.dt2tdur(dt)
    dcf = decay_const * tdur / (np.exp(-decay_const * dt[0,]) - np.exp(-decay_const * dt[1,]))
    tac_0 = tac * 1
    tac[tac < 0] = 1

    abundance = 1 / a
    prompts = abundance / dcf * tdur * tac
    normalised = prompts / np.max(prompts) * 100000000
    w = tdur * tdur / normalised / dcf
    w[tac_0 < 0] = 0
    return w
